#!/usr/bin/env python3

try:
    import sys
    import os.path
    import shutil
    import optparse
    import time
    import traceback
    import distutils.sysconfig
    import subprocess
    import socket
    import re
    import bz2
    import smtplib
    from io                     import IOBase
except ImportError as e:
    module = str(e).split()[-1]


reNVRA  = re.compile( r'(?P<name>.*)-(?P<v>[^-]+)-(?P<r>[^-]+)\.(?P<a>[^.]+)' )
reNVRAP = re.compile( r'(?P<name>.*)-(?P<v>[^-]+)-(?P<r>[^-]+)\.(?P<a>[^.]+)\.rpm' )
reNEVRA = re.compile( r'(?P<name>.*)-(?P<e>\d+):(?P<v>[^-]+)-(?P<r>[^-]+)\.(?P<a>[^.]+)' )
reNVRSP = re.compile( r'(?P<name>.*)-(?P<v>[^-]+)-(?P<r>[^-]+)\.src\.rpm' )


def getNVRAP ( path ):
    name    = None
    version = None
    release = None
    arch    = None
    directory, file = os.path.split( path )
    m = reNVRAP.match( file )
    if m:
        name    = m.group('name')
        version = m.group('v')
        release = m.group('r')
        arch    = m.group('a')
    print( name )
    return name, version, release, arch


class ErrorMessage ( Exception ):

    def __init__ ( self, code, *arguments ):
        self._code   = code
        self._errors = [ 'Malformed call to ErrorMessage()', '{}'.format(arguments) ]
        text = None
        if len(arguments) == 1:
            if isinstance(arguments[0],Exception): text = str(arguments[0]).split('\n')
            else:
                self._errors = arguments[0]
        elif len(arguments) > 1:
            text = list(arguments)
        if text:
            self._errors = []
            while len(text[0]) == 0: del text[0]
            lstrip = 0
            if text[0].startswith('[ERROR]'): lstrip = 8
            for line in text:
                if line[0:lstrip  ] == ' '*lstrip or \
                   line[0:lstrip-1] == '[ERROR]':
                    self._errors += [ line[lstrip:] ]
                else:
                    self._errors += [ line.lstrip() ]
        return

    def __str__ ( self ):
        if not isinstance(self._errors,list):
            return "[ERROR] {}".format(self._errors)
        formatted = "\n"
        for i in range(len(self._errors)):
            if i == 0: formatted += "[ERROR] {}".format(self._errors[i])
            else:      formatted += "        {}".format(self._errors[i])
            if i+1 < len(self._errors): formatted += "\n"
        return formatted

    def addMessage ( self, message ):
        if not isinstance(self._errors,list):
            self._errors = [ self._errors ]
        if isinstance(message,list):
            for line in message:
                  self._errors += [ line ]
        else:
            self._errors += [ message ]
        return

    def terminate ( self ):
        print( self )
        sys.exit(self._code)

    @property
    def code ( self ): return self._code


class BadBinary ( ErrorMessage ):

    def __init__ ( self, binary ):
        ErrorMessage.__init__( self, 1, 'Binary not found: "{}".'.format(binary) )
        return


class BadReturnCode ( ErrorMessage ):

    def __init__ ( self, status ):
        ErrorMessage.__init__( self, 1, 'Command returned status:{}.'.format(status) )
        return


class MissingSrpm ( ErrorMessage ):

    def __init__ ( self, packName ):
        ErrorMessage.__init__( self, 2, 'SRPMS "{}" not found in repositories.'.format(packName) )
        return


class UnresolvedDeps ( ErrorMessage ):

    def __init__ ( self, depends ):
        strDepends = ''
        for depend in depends:
            strDepends += ' ' + depend
        ErrorMessage.__init__( self, 3, 'Unresolved dependencies {}.'.format(strDepends) )
        return


class FailedBuild ( ErrorMessage ):

    def __init__ ( self, packName ):
        ErrorMessage.__init__( self, 4, 'Failed build for {}.'.format(packName) )
        return


class InstallFail ( ErrorMessage ):

    def __init__ ( self, rpms ):
        ErrorMessage.__init__( self, 4, 'Install fail for {}.'.format(rpms) )
        return


class Command ( object ):

    def __init__ ( self, arguments, fdLog=None, fdFilter=None ):
        self.arguments = arguments
        self.fdLog     = fdLog
        self.fdFilter  = fdFilter
        if self.fdLog != None and not isinstance(self.fdLog,IOBase):
            print( '[WARNING] Command.__init__(): "fdLog" is neither None or a file.' )
        return

    def _argumentsToStr ( self, arguments ):
        s = ''
        for argument in arguments:
            if argument.find(' ') >= 0: s += ' "' + argument + '"'
            else:                       s += ' '  + argument
        return s

    def log ( self, text ):
        line = None
        if isinstance(text,bytes):
            try:
                line = text.decode('utf-8')
            except:
                line = ''
        elif isinstance(text,str):
            line = text
        else:
            print( '[ERROR] Command.log(): "text" is neither bytes or str.' )
            print( '        {}'.format(text) )
        if line is not None:
            if isinstance(self.fdLog,IOBase):
                if self.fdFilter:
                    self.fdFilter.logFilter( line )
                try:
                    self.fdLog.write( line )
                except:
                    pass
                self.fdLog.flush()
            else:
                print( line[:-1] )
        sys.stdout.flush()
        sys.stderr.flush()
        return

    def execute ( self ):
        global conf
        sys.stdout.flush()
        sys.stderr.flush()
        
        homeDir = os.environ['HOME']
        workDir = os.getcwd()
        if homeDir.startswith(homeDir):
            workDir = '~' + workDir[ len(homeDir) : ]
        user = 'root'
        if 'USER' in os.environ: user = os.environ['USER']
        prompt = '{}@{}:{}$'.format(user,'fox',workDir)
        
        try:
            self.log( '{}{}\n'.format(prompt,self._argumentsToStr(self.arguments)) )
            #print( self.arguments )
            child = subprocess.Popen( self.arguments, stdout=subprocess.PIPE, stderr=subprocess.STDOUT )
            while True:
                line = child.stdout.readline()
                if not line: break
                self.log( line )
        except OSError as e:
            raise BadBinary( self.arguments[0] )
        
        (pid,status) = os.waitpid( child.pid, 0 )
        status >>= 8
        return status


class CommandArg ( object ):

    def __init__ ( self, command, wd=None, fdLog=None ):
        self.command = command
        self.wd      = wd
        self.fdLog   = fdLog
        return

    def __str__ ( self ):
        s = ''
        if self.wd: s = 'cd {} && '.format(self.wd)
        for i in range(len(self.command)):
            if i: s += ' '
            s += self.command[i]
        return s

    def logFilter ( self, line ):
        pass

    def getArgs ( self ):
        return self.command

    def execute ( self ):
        if self.wd: os.chdir( self.wd )
        return Command( self.getArgs(), self.fdLog, fdFilter=self ).execute()


class Rpmbuild ( CommandArg ):

    reDepends = re.compile( r'\s*(?P<depend>[\S]+).*is needed by.*' )
    rePackage = re.compile( r'^Wrote: (?P<package>[\S]+\.rpm)$' )
    reAkmods  = re.compile( r'.*/BUILDROOT/.*/akmods/.*' )

    def __init__ ( self, command, fdLog=None ):
        CommandArg.__init__ ( self, command, fdLog=fdLog )
        self.depends     = []
        self.binaries    = []
        self.debugsource = []
        self.debuginfo   = []
        self.source      = []
        return

    def logFilter ( self, line ):
        try:
            print( line[:-1] )
        except:
            pass
        m = Rpmbuild.reDepends.match( line )
        if m:
            self.depends.append( m.group('depend') )
            return
        m = Rpmbuild.rePackage.match( line )
        if not m:
            return
        mAkmods = Rpmbuild.reAkmods.match( line )
        if mAkmods:
            return
        package = m.group('package')
        if package.endswith('src.rpm'):
            self.source.append( package )
        elif package.find('debugsource') != -1:
            self.debugsource.append( package )
        elif package.find('debuginfo') != -1:
            self.debuginfo.append( package )
        else:
            self.binaries.append( package )


class Repository ( object ):

    repos = []

    @staticmethod
    def find ( packName ):
        for repo in Repository.repos:
            if packName in repo.packages:
                return repo, repo.packages[ packName ]
        return None, None

    def __init__ ( self, rootDir, disttag ):
        self.rootDir  = rootDir
        self.disttag  = disttag
        self.packages = {}
        for root, dirs, files in os.walk(rootDir):
            for file in files:
                if not file.endswith('.src.rpm'):
                    continue
                m = reNVRSP.match( file )
                if m:
                    if m.group('name') in self.packages:
                       self.packages[ m.group('name') ].append( 
                            ( root + '/' + file
                            , m.group('name')
                            , m.group('v')
                            , m.group('r')
                            ))
                    else:
                       self.packages[ m.group('name') ] \
                          = [ ( root + '/' + file
                              , m.group('name')
                              , m.group('v')
                              , m.group('r')
                              ) ]
        Repository.repos.append( self )

    def show ( self ):
        print( 'Repository "{}"'.format( self.rootDir ))
        print( 'Contains {} source packages.'.format( len(self.packages) ))
        for name, srpms in self.packages.items():
            print( '{}: {}'.format( name, srpms[0] ))



class Package ( object ):

    NO_NEVRA = 0x0001
    DONE     = 0x0002
    FORCE    = 0x0004
    NOSPEC   = 0x0008
    i686     = 0x0010

    def __init__ ( self, info ):
        self.flags = 0
        fields = info.split()
        nevra  = fields[0]
        if nevra.find(':') < 0:
            m      = reNVRA.match( nevra )
            self.E = 0
        else:
            m      = reNEVRA.match( nevra )
            self.E = m.group( 'e' )
        if m:
            self.name = m.group( 'name' )
            self.V    = m.group( 'v' )
            self.R    = m.group( 'r' )
            self.A    = m.group( 'a' )
            if self.name.endswith('-devel' ): self.name = self.name[:-6]
            if self.name.endswith('-libs'  ): self.name = self.name[:-5]
            if self.name.endswith('-common'): self.name = self.name[:-7]
            if self.name == 'python3-ipython':    self.name = 'ipython'
            if self.name == 'python3-flit-core':  self.name = 'python-flit'
            if self.name == 'scalapack-common':   self.name = 'scalapack'
            if self.name == 'vapoursynth-libs':   self.name = 'vapoursynth'
            if self.name == 'fluid-soundfont-gm': self.name = 'fluid-soundfont'
            if self.name == 'lirc-config':        self.name = 'lirc'
            if self.name == 'libshaderc':         self.name = 'shaderc'
            if self.name == 'qt6-rpm-macros':     self.name = 'qt6'
            if self.name == 'black':              self.name = 'python-black'
        else:
            self.flags |= Package.NO_NEVRA
            self.name   = nevra
        if len(fields) > 1:
            sflags = fields[1].split(',')
            for sflag in sflags:
                if sflag.upper() == 'DONE':   self.flags |= Package.DONE
                if sflag.upper() == 'FORCE':  self.flags |= Package.FORCE
                if sflag.upper() == 'NOSPEC': self.flags |= Package.NOSPEC
                if sflag.upper() == 'I686':   self.flags |= Package.i686
        self.binaries32    = []
        self.debuginfo32   = []
        self.debugsource32 = []

    def __str__ ( self ):
        if self.flags & Package.NO_NEVRA:
            s = '<{}>'.format( self.name )
        else:
            s = '<{} {} {} {} {}>'.format( self.name, self.E, self.V, self.R, self.A )
        return s

    def isDone ( self ): return self.flags & Package.DONE

    def isInstalled ( self ):
        if self.flags & Package.FORCE: return False
       #if self.name == 'avahi':        return False
       #if self.name == 'clutter':      return False
       #if self.name == 'cogl':         return False
       #if self.name == 'freeglut':     return False
       #if self.name == 'lapack':       return False
       #if self.name == 'atlas':        return False
       #if self.name == 'flexiblas':    return False
       #if self.name == 'hdf5':         return False
       #if self.name == 'liburing':     return False
       #if self.name == 'libmpeg2':     return False
       #if self.name == 'libdvdread':   return False
       #if self.name == 'libdvdnav':    return False
       #if self.name == 'fdk-aac-free': return False
       #if self.name == 'groff':        return False
       #if self.name == 'egl-wayland':  return False
       #if self.name == 'gstreamer1':   return False
       #if self.name == 'gstreamer1-plugins-base': return False
       #if self.name == 'double-conversion': return False
       #if self.name == 'postgresql': return False
       #if self.name == 'python3-aiohttp': return False
       #if self.name == 'python3-pathspec': return False
       #if self.name == 'python3-setuptools_scm': return False
        if self.name == 'glslang':      return True
       #if self.name == 'ipython':      return True
        if self.name == 'mbedtls':      return True
        if self.name == 'lirc':         return True
        if self.name == 'daala':        return True
        if self.name == 'shaderc':      return True
        if self.name == 'python-black': return True
        if self.name == 'urw-base35-fonts': return True
        dnfCheck = [ 'sudo', 'dnf', 'list', self.name ]
        status = CommandArg( dnfCheck ).execute()
        if status: return False
        return True

    def build ( self ):
        if self.isDone():
            print( '- {} flagged as DONE.'.format( self ))
            return
       #raise ErrorMessage( 0, 'Rebuild {}'.format(self) )
        print( '- Building {}.'.format( self ))
        self.rmpyegg()
        logFile = os.path.join( 'logs', self.name+'.log' )
        with open(logFile,'w') as fdLog:
            repo, origSrpms = Repository.find( self.name )
            if not origSrpms and self.name.startswith('python3'):
                self.name = self.name.replace( 'python3', 'python' )
                print( 'NEW {}'.format( self.name ))
                repo, origSrpms = Repository.find( self.name )
            if not origSrpms:
                raise MissingSrpm( self.name )
            command = [ 'rpm', '-i', origSrpms[0][0] ]
            status = Rpmbuild( command, fdLog=fdLog ).execute()
            refSpec = os.path.join( 'SPECS'    , self.name+'.spec' )
            socSpec = os.path.join( 'SPECS.SoC', self.name+'.spec' )
            if os.path.isfile(socSpec) and not self.flags & Package.NOSPEC:
                print( 'Overwritting with {}'.format( socSpec ))
                shutil.copyfile( socSpec, refSpec )
            args = [ '--clean'
                   , '--nocheck', '--without', 'tests', '--without', 'check'
                   , '--define', 'dist .{}.el9'.format(repo.disttag) ]
            if self.name == 'dbus-c++': args += [ '--without', 'ecore' ]
            if self.name == 'xfconf':   args += [ '--without', 'perl' ]
            if self.name == 'opencv':   args += [ '--without', 'clp' ]
            if self.name == 'ffmpeg':   args += [ '--without', 'frei0r'
                                                , '--without', 'x264'
                                                , '--without', 'x265'
                                                ]
            if self.name == 'texlive-base':      args += [ '--with'   , 'bootstrap' ]
            if self.name == 'qscintilla':        args += [ '--define' , 'bootstrap 1' ]
            if self.name == 'pam':               args += [ '--with'   , 'nis' ]
            if self.name == 'ghdl':              args += [ '--without', 'llvm' ]
           #if self.name.startswith('coin'):     args += [ '--with'   , 'bootstrap' ]
            if self.name in [ 'openni-primesense'
                            , 'libmpdclient'
                            , 'obs-studio'
                            , 'python-matplotlib'
                            , 'qt6-qttools'
                            , 'python-pyqt6' 
                            , 'v4l2loopback' 
                            , 'HandBrake' ] \
              or self.name.startswith('R-'):
                args += [ '--define' , 'debug_package %{nil}' ]
            args += [ refSpec ]
            command   = [ 'rpmbuild', '-ba' ] + args
            command32 = [ 'rpmbuild', '--target=i686-redhat-linux', '-bb' ] + args
            builder = Rpmbuild( command, fdLog=fdLog )
            status = builder.execute()
            if status:
                if builder.depends:
                    dnfInstall = [ 'sudo'
                                 , 'dnf', '--skip-broken', '--assumeyes', 'install' ] + builder.depends
                    CommandArg( dnfInstall ).execute()
                    dnfInstall = [ 'sudo'
                                 , 'dnf', '--assumeyes', 'install' ] + builder.depends
                    status = CommandArg( dnfInstall ).execute()
                    if status:
                        raise UnresolvedDeps( builder.depends )
                    status = builder.execute()
            if status:
                raise FailedBuild( self.name )
            if self.flags & Package.i686:
                if builder.depends:
                    depends = []
                    for depend in builder.depends:
                        depends.append( depend+'.i686' )
                    dnfInstall = [ 'sudo'
                                 , 'dnf', '--skip-broken', '--assumeyes', 'install' ] + depends
                    CommandArg( dnfInstall ).execute()
                builder32 = Rpmbuild( command32, fdLog=fdLog )
                builder32.execute()
                self.binaries32    = builder32.binaries
                self.debuginfo32   = builder32.debuginfo
                self.debugsource32 = builder32.debugsource
            self.rmsource()
            self.source      = builder.source
            self.binaries    = builder.binaries
            self.debuginfo   = builder.debuginfo
            self.debugsource = builder.debugsource
            print( 'SRPMS:' )
            for package in self.source:
                print( '- {}'.format( package ))
            print( 'RPMS:' )
            for package in self.binaries + self.binaries32:
                print( '- {}'.format( package ))
            print( 'debuginfo:' )
            for package in self.debuginfo + self.debuginfo32:
                print( '- {}'.format( package ))
            print( 'debugsource:' )
            for package in self.debugsource + self.debugsource32:
                print( '- {}'.format( package ))
            rpmsign = [ 'rpm', '--addsign' ] + self.source        \
                                             + self.binaries      \
                                             + self.binaries32    \
                                             + self.debuginfo     \
                                             + self.debuginfo32   \
                                             + self.debugsource   \
                                             + self.debugsource32
            CommandArg( rpmsign ).execute()
            self.installNames = []
            for path in self.binaries:
                name, v, r, a = getNVRAP( path )
                if name:
                    self.installNames.append( name )

    def install ( self ):
        dnfInstall = [ 'sudo'
                     , 'dnf'
                     , '--assumeyes', 'install' ] + self.installNames
        status = CommandArg( dnfInstall ).execute()
        if status:
            raise InstallFail( self.binaries )

    def rmpyegg ( self ):
        for extension in [ 'files', 'modules', 'record', 'ghost-distinfo' ]:
            pyfile = 'BUILD/pyproject-' + extension
            if os.path.isfile(pyfile):
                os.unlink( pyfile )

    def rmsource ( self ):
        refSpec = os.path.join( 'SPECS', self.name+'.spec' )
        command = [ 'rpmbuild', '--rmsource', refSpec ]
        Rpmbuild( command ).execute()

    def toRepository ( self, repoDir, packages ):
        for package in packages:
            directory, file = os.path.split( package )
            shutil.move( package, repoDir+'/Packages/'+file, copy_function=os.replace )
        createrepo = [ 'createrepo_c', '-d', repoDir ]
        CommandArg( createrepo ).execute()

    def toRepositories ( self, repoDirRoot ):
        self.toRepository( repoDirRoot + '/source'      , self.source )
        self.toRepository( repoDirRoot + '/x86_64/os'   , self.binaries )
        self.toRepository( repoDirRoot + '/x86_64/debug', self.debuginfo + self.debugsource )
        CommandArg( [         'dnf', 'clean', 'all' ] ).execute()
        CommandArg( [ 'sudo', 'dnf', 'clean', 'all' ] ).execute()


if __name__ == '__main__':
    distribDir  = '/dsk/l2/distributions'
    hostname = socket.gethostname()
    if hostname.startswith('lepka'):
       #distribDir = '/run/media/jpc/SanDisk_JPC/distributions'
        distribDir = '/dsk/l1/distributions'
    rpmbuildDir = distribDir  + '/soc/9-alma/rpmbuild'
    addonsDir   = distribDir  + '/soc/9-alma/addons'
    texliveDir  = distribDir  + '/soc/9-alma/texlive'
    fedoraDir   = distribDir  + '/fedora/releases/35/Everything/source'
   #fusionDir   = distribDir  + '/rpmfusion/free/el/updates/8'
    fusionDir   = distribDir  + '/rpmfusion/free/fedora/releases/35/Everything/source'
    customDir   = rpmbuildDir + '/SRPMS.orig'
    fc35Dir     = rpmbuildDir + '/SRPMS.fc35'
    fc36Dir     = rpmbuildDir + '/SRPMS.fc36'
    fc37Dir     = rpmbuildDir + '/SRPMS.fc37'
    fc38Dir     = rpmbuildDir + '/SRPMS.fc38'
    fc39Dir     = rpmbuildDir + '/SRPMS.fc39'
    al9Dir      = rpmbuildDir + '/SRPMS.al9'
    suseDir     = rpmbuildDir + '/SRPMS.suse'
    customRepo  = Repository( customDir, 'soc' )
    al9Repo     = Repository( al9Dir   , 'al9' )
    fc39Repo    = Repository( fc39Dir  , 'fc39' )
    fc38Repo    = Repository( fc38Dir  , 'fc38' )
    fc37Repo    = Repository( fc37Dir  , 'fc37' )
    fc36Repo    = Repository( fc36Dir  , 'fc36' )
    fc35Repo    = Repository( fc35Dir  , 'fc35' )
    fedoraRepo  = Repository( fedoraDir, 'fc35' )
    fusionRepo  = Repository( fusionDir, 'fusion' )
    suseRepo    = Repository( suseDir  , 'suse' )
    customRepo.show()
    CommandArg( [         'dnf', 'clean', 'all' ] ).execute()
    CommandArg( [ 'sudo', 'dnf', 'clean', 'all' ] ).execute()
    packageList = 'packages-soc.txt'
    enabled     = False
    with open(packageList,'r') as fd:
        packageInfo = fd.readline().strip()
        while packageInfo:
            if packageInfo[0] == '#':
                packageInfo = fd.readline().strip()
                continue
            if not enabled:
                if packageInfo.startswith('ENABLED'):
                    enabled = True
                else:
                    print( '- Skip {}'.format( packageInfo ))
            else:
                #print( '{}: {}'.format( count, packName ))
                package = Package( packageInfo ) 
                if (   not package.name.endswith('-devel')  \
                   and not package.name.endswith('-common') \
                   and not package.name.endswith('-doc')    \
                   and not package.name.endswith('-about')):
                    if not package.isDone() and not package.isInstalled():
                        repoDir = addonsDir
                        if    package.name.startswith('texlive') \
                           or package.name.startswith('potrace') \
                           or package.name.startswith('t1lib')   \
                           or package.name.startswith('teckit'):
                            repoDir = texliveDir
                        package.build()
                        package.toRepositories( repoDir )
                        package.install()
            packageInfo = fd.readline().strip()
